library nueli.auth.firebase;

import 'dart:async';
import 'dart:convert';

import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:firebase_auth/firebase_auth.dart' as firebase;
import 'package:firebase_core/firebase_core.dart';
import 'package:flutter/cupertino.dart';
import 'package:nueli/nueli.dart';
import 'package:shared_preferences/shared_preferences.dart';

part 'account-link.dart';
part 'account.dart';
part 'meta.dart';

class FirebaseAuth {
  static Stream<Account> _signInChanges;
  static Stream<Account> _idTokenChanges;
  static Stream<Account> _accountChanges;

  static StreamSubscription<Account> _signInChangesSub; // ignore: cancel_subscriptions
  static StreamSubscription<Account> _idTokenChangesSub; // ignore: cancel_subscriptions
  static StreamSubscription<Account> _accountChangesSub; // ignore: cancel_subscriptions

  static List<AccountLink> _accounts;
  static AccountLink _currentAccount;
  static Account _signedAccount;

  static firebase.FirebaseAuth get _instance => firebase.FirebaseAuth.instance;

  static Future<void> _saveAccount(AccountLink account) async {
    await asyncHandle(() async {
      await SharedPreferences.getInstance().then((prefs) async {
        if (!_accounts.contains(account)) _accounts = _accounts..add(account);
        await prefs.setStringList(_Keys.accounts, _accounts.map((e) => e.compile().compile()).toList());
        await prefs.setString(_Keys.currentAccount, account.compile().compile());
      });
    });
  }

  static Future<void> initialize() async {
    await SharedPreferences.getInstance().then((value) {
      _accounts = [];
      if (value.containsKey(_Keys.accounts)) // break
        _accounts = value.getStringList(_Keys.accounts).map((e) => AccountLink.parse(e.parseAsMap())).toList();
      if (value.containsKey(_Keys.currentAccount)) // break
        _currentAccount = AccountLink.parse(value.getString(_Keys.currentAccount).parseAsMap());
    });
    _signInChanges = (_instance.authStateChanges()..listen((user) => // break
        _signedAccount = (user == null) ? null : Account.from(user: user))).map((user) => Account.from(user: user));
    _idTokenChanges = (_instance.idTokenChanges()..listen((user) => // break
        _signedAccount = (user == null) ? null : Account.from(user: user))).map((user) => Account.from(user: user));
    _accountChanges = (_instance.userChanges()..listen((user) => // break
        _signedAccount = (user == null) ? null : Account.from(user: user))).map((user) => Account.from(user: user));
  }

  static Future<void> signIn({
    @required String email,
    @required String password,
    bool savePassword = false,
  }) async {
    /// User Credential
    firebase.UserCredential user;

    throwException(FirebaseException e) {
      String error = e.code;
      if (e.code == 'user-not-found')
        error = glossary.invalidAccount;
      else if (e.code == 'wrong-password')
        error = glossary.incorrectPassword;
      else if (e.code == 'too-many-requests')
        error = glossary.overRequest;
      else if (e.code == 'network-request-failed') error = glossary.badNetwork;
      throw Exception(error);
    }

    try {
      user = await _instance.signInWithEmailAndPassword(email: email, password: password);
      if (!user.user.emailVerified) throw Exception(glossary.verifyEmail);
    } on firebase.FirebaseAuthException catch (e) {
      throwException(e);
    }

    _signedAccount = Account._(
      name: user.user.displayName,
      email: email,
      avatarLink: user.user.photoURL,
      creationTime: user.user.metadata.creationTime,
      phoneNumber: user.user.phoneNumber,
      id: user.user.uid,
      lastSignIn: user.user.metadata.lastSignInTime,
    )..user = user;
    _currentAccount = AccountLink.from(_signedAccount, password: savePassword ? password : '');
    _saveAccount(_currentAccount);
    return _signedAccount;
  }

  static Future<Account> signUp({
    @required String email,
    @required String password,
    @required String name,
  }) async {
    /// User Credentials
    firebase.UserCredential user;

    throwException(FirebaseException e) {
      String error = e.code;
      if (e.code == 'weak-password') // break
        error = glossary.weakPassword;
      else if (e.code == 'email-already-in-use') // break
        error = glossary.usedEmail;
      else if (e.code == 'too-many-requests') // break
        error = glossary.overRequest;
      else if (e.code == 'unknown') // break
        error = glossary.unknown;
      throw Exception(error);
    }

    if (email == null || email.isEmpty) throw Exception(glossary.needEmail);
    if (password == null && password.length < 8) throw Exception(glossary.invalidPassword);

    try {
      user = await FirebaseAuth._instance.createUserWithEmailAndPassword(email: email, password: password);
      await user.user.updateDisplayName(name);
      await user.user.sendEmailVerification();
    } on firebase.FirebaseAuthException catch (e) {
      throwException(e);
    }

    final account = Account._(
      name: name,
      email: user.user.email,
      phoneNumber: user.user.phoneNumber,
      id: user.user.uid,
      lastSignIn: user.user.metadata.lastSignInTime,
      creationTime: user.user.metadata.creationTime,
    )..user = user;
    _currentAccount = AccountLink.from(account);
    return account;
  }

  static Future<void> signOut() async {
    await _instance.signOut();
    _signedAccount = null;
  }

  static AccountLink get currentAccount => _currentAccount;
  static bool get isSignedIn => _signedAccount != null;
  static Account get signedAccount => _signedAccount ?? null;

  static bool get hasAccounts => _accounts.isNotEmpty;
  StreamSubscription<Account> get signInChanges => _signInChangesSub;
  StreamSubscription<Account> get idTokenChanges => _idTokenChangesSub;
  StreamSubscription<Account> get accountsChanges => _accountChangesSub;

  void cancelSubscription() {
    if (signInChanges != null) signInChanges.cancel();
    if (idTokenChanges != null) idTokenChanges.cancel();
    if (accountsChanges != null) accountsChanges.cancel();
  }

  static void subscribeTo({
    Function(Account signedUser) signInChanges,
    Function(Account signedUser) idTokenChanges,
    Function(Account signedUser) accountChanges,
  }) {
    if (signInChanges != null) _signInChangesSub = _signInChanges.listen((account) => signInChanges(account));
    if (idTokenChanges != null) _idTokenChangesSub = _idTokenChanges.listen((account) => idTokenChanges(account));
    if (accountChanges != null) _accountChangesSub = _accountChanges.listen((account) => accountChanges(account));
  }
}
