part of nueli.auth.firebase;

class Account {
  final String id;
  final String email;
  final String phoneNumber;
  final DateTime lastSignIn;
  final String name;
  final DateTime creation;
  final String avatarLink;
  firebase.UserCredential user;

  // constructors
  Account._({
    this.id = '',
    this.email = '',
    this.phoneNumber = '',
    String name = '',
    this.avatarLink = '',
    DateTime lastSignIn,
    DateTime creationTime,
  })  : name = name.replaceAll('  ', ' '),
        lastSignIn = lastSignIn ?? DateTime.now(),
        creation = creationTime ?? DateTime.now();

  factory Account.from({@required firebase.User user}) => Account._(
        email: user.email,
        name: user.displayName,
        phoneNumber: user.phoneNumber,
        id: user.uid,
        lastSignIn: user.metadata.lastSignInTime,
        creationTime: user.metadata.creationTime,
        avatarLink: user.photoURL,
      );

  factory Account.parse(Map<String, String> data) {
    if (!data.containsKey(_Keys.id)) if (!data.containsKey(_Keys.passcode)) // break
    if (!data.containsKey(_Keys.name)) if (!data.containsKey(_Keys.lastSignIn)) // break
    if (!data.containsKey(_Keys.email)) if (!data.containsKey(_Keys.flags)) // break
    if (!data.containsKey(_Keys.phoneNumber)) if (!data.containsKey(_Keys.creation)) // break
    if (!data.containsKey(_Keys.avatarLink)) throw Exception(glossary.fatalError);

    return data != null
        ? Account._(
            id: data[_Keys.id],
            name: data[_Keys.name],
            email: data[_Keys.email],
            phoneNumber: data[_Keys.phoneNumber],
            lastSignIn: DateTime.tryParse(data[_Keys.lastSignIn]) ?? DateTime.now(),
            creationTime: DateTime.tryParse(data[_Keys.creation]) ?? DateTime.now(),
            avatarLink: data[_Keys.avatarLink],
          )
        : Account._();
  }

  // private static
  Future<CollectionReference> get _detailRef async {
    await Firebase.initializeApp();
    return FirebaseFirestore.instance.collection(_Keys.meta);
  }

  // static
  static Future<Account> anomalous() async {
    bool successful = true;
    firebase.UserCredential user = await FirebaseAuth._instance.signInAnonymously().catchError((error) {
      successful = false;
    });
    if (!successful) throw Exception('An error occurred');
    return Account._(
      id: '@anomalous',
      lastSignIn: DateTime.now(),
    )..user = user;
  }

  // instance
  bool get isVerified => user.user.emailVerified;

  String get firstName => handle(() => name.split(' ').first, alt: (e) => '');
  String get lastName => handle(() => name.split(' ').last, alt: (e) => '');

  Future<void> setDetail(String key, String Function(String former) detail) async {
    final doc = (await _detailRef).doc(id);
    await asyncHandle(() async {
      if (!((await doc.get()).exists)) await doc.set({});
      final data = Map<String, String>.from((await doc.get()).data());
      doc.set(data..update(key, (value) => detail(value), ifAbsent: () => detail(null)));
    });
  }

  Future<String> getDetail(key) async {
    final doc = (await _detailRef).doc(id);
    return await asyncHandle(() async {
      if (!((await doc.get()).exists)) await doc.set({});
      final data = Map<String, String>.from((await doc.get()).data());
      if (data.containsKey(key)) return data[key];
      return null;
    });
  }

  Future<void> resendVerificationEmail() async => await user.user.sendEmailVerification();

  login(String password) {
    firebase.FirebaseAuth.instance.signInWithEmailAndPassword(email: email, password: password);
  }

  AccountLink buildLink([String password]) => AccountLink.from(this, password: password);

  Map<String, String> compile() {
    return {
      _Keys.id: id,
      _Keys.name: name,
      _Keys.email: email,
      _Keys.avatarLink: avatarLink,
      _Keys.phoneNumber: phoneNumber,
      _Keys.lastSignIn: (lastSignIn ?? DateTime.now()).toIso8601String(),
    };
  }
}
